#ifndef simplicialcomplex_h_nealaezeojeeChuh
#define simplicialcomplex_h_nealaezeojeeChuh

#include <stdint.h>
#include <cstdlib>
#include <cstring>
#include <algorithm>
#include <vector>
#include <limits>

#include <iostream>

#include "booleanmatrix.h"


namespace libstick {

/** A simplicial complex is a collection of simplices such that each face is
 * also part of the complex. Every simplex has dimension at most MAXDIM. The
 * indices of simplices and their faces are of type IT. When a
 * simplicial_complex is instantiated, a single (-1) dimensional simplex is
 * automatically created. (When we compute persistent homology, we actually
 * compute reduced homology.) Each 0-dimensional simplex automatically has this
 * simplex as its face.
 *
 * Based on a simplicial complex, we define a simplicial order. An order
 * can be seen as a permutation of the simplices of the complex. If the order
 * has the property that every prefix of the permutation is again a complex,
 * i.e., all faces of all simplices of each prefix are contained in the prefix,
 * then we call it a filtration. The innner class simplex_order gives the
 * extended boundary matrix, which can then be used in the persistence class to
 * compute persistent homology.
 *
 * One way to obtain a filtration is to define a monotone simplicial function
 * on the complex. That is, each simplex gets a value of type VT assigned. Then
 * one can obtain a sub-level set filtration from the complex w.r.t. to this
 * simplicial function.
 */
template<int MAXDIM, class IT>
class simplicial_complex {

    public:
        /** The type of this class. */
        typedef simplicial_complex<MAXDIM, IT> simplcompltype;
        /** Type of indices of simplices. */
        typedef IT index_type;

        /** A simplex of the complex. This struct is an aggregate. */
        struct simplex {
            /** Dimension of the simplex. */
            int dim;
            /** The indices of the faces of the simplex. */
            index_type faces[MAXDIM+1];

            /** Create a new simplex with dimension 'dim' and (dim+1)-faces. If
             * the simplex is 0-dimensional, its face is automatically set to
             * one (-1)-dimensional simplex. */
            static simplex create_simplex(int dim, index_type* faces) {
                assert(0 <= dim && dim <= MAXDIM);

                simplex s = {dim, {0}};

                if (dim >= 1)
                    memcpy(s.faces, faces, face_count_bydim(dim)*sizeof(index_type));

                return s;
            }

            /** Create a (-1)-dimensional simplex. */
            static simplex create_minusonedim_simplex() {
                simplex s = {-1, {0}};
                return s;
            }

            /** Get number of faces. */
            size_t face_count() const {
                return face_count_bydim(dim);
            }

            /** Get number of faces of a dim-dimensional simplex. */
            static size_t face_count_bydim(int dim) {
                assert(-1 <= dim && dim <= MAXDIM);
                return dim + 1;
            }
        };

        /** An order of the simplices of complex c. An order can be interpreted
         * as a permuation of the complex's std::vector of simplices.  */
        class simplex_order {

            public:
                typedef boolean_colmatrix<IT> boundary_matrix;
                typedef std::vector<IT> order_type;

                /** Create a standard order of the complex c, i.e., the identity permutation. */
                simplex_order(const simplcompltype &c) :
                    c(c) {
                    reset();
                }

                /** Reset order to the identity permutation of the complex's simplices. */
                void reset() {
                    order.clear();
                    for (unsigned i=0; i < c.size(); ++i)
                        order.push_back(i);
                    revorder = order;
                }

                /** Return number of simplices. */
                size_t size() const {
                    assert(order.size() == revorder.size());
                    return order.size();
                }

                /** Get i-th simplex in the simplex order. */
                const simplex& get_simplex(index_type i) const {
                    return c.simplices.at(order_to_complex(i));
                }

                /** Translate an index in the order to a simplex in the complex. */
                index_type order_to_complex(index_type i) const {
                    assert(0 <= i && i < size());
                    return order.at(i);
                }

                /** Translate an index in the complex to a simplex in the order. */
                index_type complex_to_order(index_type i) const {
                    assert(0 <= i && i < size());
                    return revorder.at(i);
                }

                /** Return the underlying complex. */
                const simplcompltype& get_complex() const {
                    return c;
                }

                /** Check if this order is a permutation of the underlying
                 * complex. */
                bool is_permutation() const {
                    if (size() != c.size())
                        return false;

                    // Sort the index array and check if every index is present.
                    order_type o = order;
                    std::sort(o.begin(), o.end());
                    for (unsigned i=0; i < size(); ++i)
                        if (o.at(i) != i)
                            return false;

                    return true;
                }

                /** Returns true iff the faces of simplex i are before i in
                 * this order. */
                bool is_filtration() const {
                    assert(size() == c.size());

                    for (unsigned i=0; i < size(); ++i)
                        for (unsigned f=0; f < get_simplex(i).face_count(); ++f)
                            if (complex_to_order(get_simplex(i).faces[f]) >= i)
                                return false;

                    return true;
                }

                /** Randomize order. It has hardly any impact on runtime, but
                 * it makes cycles "nicer" when then making order monotone
                 * w.r.t. a simplicial function. */
                void randomize_order() {
                    std::random_shuffle(order.begin(), order.end());
                    restore_revorder_from_order();
                }

                /** Get the boundary matrix of the complex according to this
                 * order. Requires that is_permutation() gives true. */
                boundary_matrix get_boundary_matrix() const {
                    assert(is_permutation());
                    boundary_matrix mat(size());

                    for (unsigned c=0; c < size(); ++c)
                        for(unsigned r=0; r < get_simplex(c).face_count(); ++r)
                            mat.set(complex_to_order(get_simplex(c).faces[r]), c, true);

                    return mat;
                }

                /** Get the order of the simplices. */
                const order_type& get_order() const {
                    return order;
                }

                /** Set the order of the simplices. Requires that order has the
                 * same size as the underlying complex. Requires that the
                 * is_permutation() gives true after setting the order. */
                void set_order(const order_type& o) {
                    assert(o.size() == c.size());

                    order = o;
                    restore_revorder_from_order();
                }

            private:
                /** Reconstruct 'revorder' by inverting the permutation given by 'order'. */
                void restore_revorder_from_order() {
                    assert(is_permutation());

                    // Make revorder * order the identity permutation
                    for (unsigned i=0; i < size(); ++i)
                        revorder[order[i]] = i;
                }

                /** The complex of which we consider a simplex order. */
                const simplcompltype &c;

                /** The i-th simplex in order is the order[i]-th simplex of the
                 * complex. 'order' can be seen as a permutation of the
                 * simplices saved in 'c'. */
                order_type order;

                /** The i-th simplex in the complex is the revorder[i]-th
                 * simplex in order. 'revorder' can be seen as the inverse
                 * permutation saved in 'order'. */
                order_type revorder;
        };

    public:
        simplicial_complex() {
            // Add the one and only minus-one dimensional simplex
            simplices.push_back(simplex::create_minusonedim_simplex());
        }

        /** Remove all simplices except the dummy simplex */
        void clear() {
            simplices.resize(1);
        }

        /** Return number of simplices. */
        size_t size() const {
            return simplices.size();
        }

        /** Add a simplex to the complex. The dimension of the faces must be
         * dim-1, and they must already be part of the complex. Returns the
         * index of the added simplex. */
        index_type add_simplex(int dim, index_type* faces) {
            return add_simplex(simplex(dim, faces));
        }

        /** Add a simplex to the complex. The dimension of the faces must be
         * dim-1, and they must already be part of the complex. Returns the
         * index of the added simplex. The dimension must be at least zero. */
        index_type add_simplex(simplex s) {
            assert(s.dim >= 0);

            // Check requirements for faces
            for (unsigned i=0; i < s.face_count(); ++i) {
                // Faces are already in complex.
                assert(s.faces[i] < size());
                // Faces have dimension dim-1
                assert(simplices[s.faces[i]].dim == s.dim-1);
            }

            // index_type must be large enough
            assert(simplices.size() < std::numeric_limits<IT>::max());

            index_type idx = simplices.size();
            simplices.push_back(s);
            return idx;
        }

        /** Add an array of simplices */
        void add_simplices(simplex* sarray, size_t count) {
            for (unsigned i=0; i < count; ++i)
                add_simplex(sarray[i]);
        }

        /** Get idx-th simplex */
        const simplex& get_simplex(index_type idx) const {
            assert(0 <= idx && idx < size());
            return simplices.at(idx);
        }

        /** Return true iff for each simplex i with dimension dim it holds that
         * the faces of i are contained in the complex and have dimension dim-1. */
        bool is_complex() const {
            for (unsigned i=0; i < size(); ++i) {

                const simplex &s = simplices[i];
                for (unsigned f=0; f < s.face_count(); ++f) {

                    if (s.faces[f] >= size())
                        return false;

                    const simplex &face = simplices[s.faces[f]];
                    if (face.dim != s.dim-1)
                        return false;
                }
            }
            return true;
        }

    private:
        /** A list of simplices */
        std::vector<simplex> simplices;
};

}


#endif
